-------------------------------------------
--- Author: Ketho (EU-Boulderfist)		---
--- Created: 2009.09.01					---
--- License: Public Domain				---
-------------------------------------------
--- Ketho CombatLog	v0.98  [2011.06.29]	---
--- Ace3			r1032  [2011.06.29]	---
--- LibSink-2.0		r78    [2011.06.28]	---
-------------------------------------------
--- Curse			http://wow.curse.com/downloads/wow-addons/details/ketho-combatlog.aspx
--- WoWInterface	http://www.wowinterface.com/downloads/info18901-KethoCombatLog.html

--- I don't have any previous programming experience or background in any (scripting) language
--- Note: This is my first AddOn

--- To Do:
-- DevTools style debugging, and similar way to easily announce a chatlog entry

KethoCombatLog = LibStub("AceAddon-3.0"):NewAddon("KethoCombatLog", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0", "LibSink-2.0")
local KCL = KethoCombatLog
local ACR = LibStub("AceConfigRegistry-3.0")
local ACD = LibStub("AceConfigDialog-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("KethoCombatLogLocale", true)

	--------------------
	--- 01  UPVALUES ---
	--------------------

local _G = _G
--# Lua APIs
local pairs, type = pairs, type
local tonumber = tonumber
local unpack, select = unpack, select
local time = time
local format = format
local strsub = strsub
local strlower = strlower
local bit_band = bit.band -- BitLib library

--# WoW APIs
local UnitName, UnitClass = UnitName, UnitClass
local UnitInParty, UnitInRaid = UnitInParty, UnitInRaid
local UnitGroupRolesAssigned = UnitGroupRolesAssigned
local GetSpellInfo, GetSpellLink = GetSpellInfo, GetSpellLink

--# Global Strings / Format Strings / Bitmasks
local LOCALIZED_CLASS_NAMES_MALE = LOCALIZED_CLASS_NAMES_MALE

local TEXT_MODE_A_STRING_SOURCE_UNIT = TEXT_MODE_A_STRING_SOURCE_UNIT
local TEXT_MODE_A_STRING_DEST_UNIT = TEXT_MODE_A_STRING_DEST_UNIT
local TEXT_MODE_A_STRING_SOURCE_ICON = TEXT_MODE_A_STRING_SOURCE_ICON
local TEXT_MODE_A_STRING_DEST_ICON = TEXT_MODE_A_STRING_DEST_ICON
local TEXT_MODE_A_STRING_SPELL = TEXT_MODE_A_STRING_SPELL

local TEXT_MODE_A_STRING_RESULT_OVERKILLING = TEXT_MODE_A_STRING_RESULT_OVERKILLING
local TEXT_MODE_A_STRING_RESULT_CRITICAL = TEXT_MODE_A_STRING_RESULT_CRITICAL
local TEXT_MODE_A_STRING_RESULT_GLANCING = TEXT_MODE_A_STRING_RESULT_GLANCING
local TEXT_MODE_A_STRING_RESULT_CRUSHING = TEXT_MODE_A_STRING_RESULT_CRUSHING

local COMBATLOG_OBJECT_REACTION_FRIENDLY = COMBATLOG_OBJECT_REACTION_FRIENDLY
local COMBATLOG_OBJECT_REACTION_HOSTILE = COMBATLOG_OBJECT_REACTION_HOSTILE

local COMBATLOG_OBJECT_RAIDTARGET_MASK = COMBATLOG_OBJECT_RAIDTARGET_MASK
local COMBATLOG_OBJECT_RAIDTARGET1 = COMBATLOG_OBJECT_RAIDTARGET1
local COMBATLOG_OBJECT_RAIDTARGET2 = COMBATLOG_OBJECT_RAIDTARGET2
local COMBATLOG_OBJECT_RAIDTARGET3 = COMBATLOG_OBJECT_RAIDTARGET3
local COMBATLOG_OBJECT_RAIDTARGET4 = COMBATLOG_OBJECT_RAIDTARGET4
local COMBATLOG_OBJECT_RAIDTARGET5 = COMBATLOG_OBJECT_RAIDTARGET5
local COMBATLOG_OBJECT_RAIDTARGET6 = COMBATLOG_OBJECT_RAIDTARGET6
local COMBATLOG_OBJECT_RAIDTARGET7 = COMBATLOG_OBJECT_RAIDTARGET7
local COMBATLOG_OBJECT_RAIDTARGET8 = COMBATLOG_OBJECT_RAIDTARGET8

-- table reference shortcuts
local profile, char
local color

local filterSelf, filterEverythingElse, filterPlayers, filterMonsters
local chatWindow, chatChannel
local iconSize, Timestamp
local overkillFormat, criticalFormat, glancingFormat, crushingFormat

-- modified globalstrings
local ADVANCED = gsub(ADVANCED_LABEL, "|T.-|t", "")
local OVERKILLING = gsub(TEXT_MODE_A_STRING_RESULT_OVERKILLING, "[%%d ()]", "")
local CRITICAL = gsub(TEXT_MODE_A_STRING_RESULT_CRITICAL, "[()]", "")
local GLANCING = gsub(TEXT_MODE_A_STRING_RESULT_GLANCING, "[()]", "")
local CRUSHING = gsub(TEXT_MODE_A_STRING_RESULT_CRUSHING, "[()]", "")

-- other
local ChatFrame
local optionsFrame
local cropped = "64:64:4:60:4:60"
local _

-- arrays for colors
local SpellSchoolColor
local ClassColor
local UnitColor
local EventColor

-- internal cooldowns
local cd = {}

local player = {
	name = UnitName("player"),
	class = select(2, UnitClass("player")),
}

local Taunt = {
-- Death Knight
	[56222] = true, -- Dark Command
	[49560] = true, -- Death Grip
	[51399] = true, -- Death Grip (melee range)
-- Druid
	[6795] = true, -- Growl
-- Hunter
	[20736] = true, -- Distracting Shot
-- Paladin
	[62124] = true, -- Hand of Reckoning
	[31790] = true, -- Righteous Defense
-- Warlock	
	[17735] = true, -- Suffering [Voidwalker]
-- Warrior
	[355] = true, -- Taunt
}

local Interrupt = {
	[47528] = true, -- Mind Freeze [Death Knight]
	[80964] = true, -- Skull Bash (Bear) [Druid]
	[80965] = true, -- Skull Bash (Cat) [Druid]
	[93985] = true, -- Skull Bash (?) [Druid]
	[2139] = true, -- Counterspell [Mage]
	[96231] = true, -- Rebuke [Paladin]
	[1766] = true, -- Kick [Rogue]
	[57994] = true, -- Wind Shear [Shaman]
	[6552] = true, -- Pummel [Warrior]
}

local CrowdControl = {
-- Druid
	[2637] = true, -- Hibernate
	[33786] = true, -- Cyclone
-- Hunter
	[3355] = true, -- Freezing Trap
	[1513] = true, -- Scare Beast
	[19386] = true, -- Wyvern Sting
-- Mage
	[118] = true, -- Polymorph
	[61305] = true, -- Polymorph - Black Cat
	[28272] = true, -- Polymorph - Pig
	[61721] = true, -- Polymorph - Rabbit
	[61780] = true, -- Polymorph - Turkey
	[28271] = true, -- Polymorph - Turtle
-- Paladin
	[10326] = true, -- Turn Evil
	[20066] = true, -- Repentance
-- Priest
	[9484] = true, -- Shackle Undead
-- Rogue
	[6770] = true, -- Sap
	[2094] = true, -- Blind
-- Shaman
	[51514] = true, -- Hex
	[76780] = true, -- Bind Elemental
-- Warlock
	[710] = true, -- Banish
	[5782] = true, -- Fear
	[6358] = true, -- Seduction [Succubus]
}

local SpellSchoolString = {
	[0x1] = STRING_SCHOOL_PHYSICAL,
	[0x2] = STRING_SCHOOL_HOLY,
	[0x4] = STRING_SCHOOL_FIRE,
	[0x8] = STRING_SCHOOL_NATURE,
	[0x10] = STRING_SCHOOL_FROST,
	[0x20] = STRING_SCHOOL_SHADOW,
	[0x40] = STRING_SCHOOL_ARCANE,
	[0x3] = STRING_SCHOOL_HOLYSTRIKE,
	[0x5] = STRING_SCHOOL_FLAMESTRIKE,
	[0x6] = STRING_SCHOOL_HOLYFIRE,
	[0x9] = STRING_SCHOOL_STORMSTRIKE,
	[0xA] = STRING_SCHOOL_HOLYSTORM,
	[0xC] = STRING_SCHOOL_FIRESTORM,
	[0x11] = STRING_SCHOOL_FROSTSTRIKE,
	[0x12] = STRING_SCHOOL_HOLYFROST,
	[0x14] = STRING_SCHOOL_FROSTFIRE,
	[0x18] = STRING_SCHOOL_FROSTSTORM,
	[0x21] = STRING_SCHOOL_SHADOWSTRIKE,
	[0x22] = STRING_SCHOOL_SHADOWLIGHT, -- Twilight
	[0x24] = STRING_SCHOOL_SHADOWFLAME,
	[0x28] = STRING_SCHOOL_SHADOWSTORM, -- Plague
	[0x30] = STRING_SCHOOL_SHADOWFROST,
	[0x41] = STRING_SCHOOL_SPELLSTRIKE,
	[0x42] = STRING_SCHOOL_DIVINE,
	[0x44] = STRING_SCHOOL_SPELLFIRE,
	[0x48] = STRING_SCHOOL_SPELLSTORM,
	[0x50] = STRING_SCHOOL_SPELLFROST,
	[0x60] = STRING_SCHOOL_SPELLSHADOW,
	[0x1C] = STRING_SCHOOL_ELEMENTAL,
	[0x7C] = STRING_SCHOOL_CHROMATIC,
	[0x7E] = STRING_SCHOOL_MAGIC,
	[0x7F] = STRING_SCHOOL_CHAOS,
}

local STRING_REACTION_ICON = {
	TEXT_MODE_A_STRING_SOURCE_ICON,
	TEXT_MODE_A_STRING_DEST_ICON,
}

local extraSpellEvent = {
	["SPELL_INTERRUPT"] = true,
	["SPELL_DISPEL"] = true,
	["SPELL_DISPEL_FAILED"] = true,
	["SPELL_STOLEN"] = true,
	["SPELL_AURA_BROKEN_SPELL"] = true,
}

local missType = {
	["ABSORB"] = COMBAT_TEXT_ABSORB,
	["BLOCK"] = COMBAT_TEXT_BLOCK,
	["DEFLECT"] = COMBAT_TEXT_DEFLECT,
	["DODGE"] = COMBAT_TEXT_DODGE,
	["EVADE"] = COMBAT_TEXT_EVADE,
	["IMMUNE"] = COMBAT_TEXT_IMMUNE,
	["MISS"] = COMBAT_TEXT_MISS,
	["PARRY"] = COMBAT_TEXT_PARRY,
	["REFLECT"] = COMBAT_TEXT_REFLECT,
	["RESIST"] = COMBAT_TEXT_RESIST,
}

local environmentalDamageType = {
	["DROWNING"] = ACTION_ENVIRONMENTAL_DAMAGE_DROWNING,
	["FALLING"] = ACTION_ENVIRONMENTAL_DAMAGE_FALLING,
	["FATIGUE"] = ACTION_ENVIRONMENTAL_DAMAGE_FATIGUE,
	["FIRE"] = ACTION_ENVIRONMENTAL_DAMAGE_FIRE,
	["LAVA"] = ACTION_ENVIRONMENTAL_DAMAGE_LAVA,
	["SLIME"] = ACTION_ENVIRONMENTAL_DAMAGE_SLIME,
}

	------------------------------
	--- 02  Options & Defaults ---
	------------------------------

local defaults = {
	profile = { -- KethoCombatLog.db.defaults.profile
		enableAddOn = true,
		Taunt = true,
		Interrupt = true,
		Death = true,
		PvE = true,
		chatWindow = 2, -- ChatFrame1
		chatChannel = 1, -- Disabled
		color = {
			Taunt = {1, 0, 0}, -- #FF0000 (Red)
			Interrupt = {0, 110/255, 1}, -- #006EFF
			Dispel = {1, 1, 1},
			Reflect = {1, 1, 1},
			CrowdControl = {1, 1, 1},
			CC_Break = {1, 1, 1},
			Death = {1, 1, 1},
			Resurrection = {175/255, 1, 47/255}, -- #ADFF2F (GreenYellow)
			-- Constants.lua: SCHOOL_MASK_PHYSICAL, ...
			Physical = {1.00, 1.00, 0.00}, -- #FFFF00
			Holy = {1.00, 0.90, 0.50}, -- #FFE680
			Fire = {1.00, 0.50, 0.00}, -- #FF8000
			Nature = {0.30, 1.00, 0.30}, -- #4DFF4D
			Frost = {0.50, 1.00, 1.00}, -- #80FFFF
			Shadow = {0.50, 0.50, 1.00}, -- #8080FF
			Arcane = {1.00, 0.50, 1.00}, -- #FF80FF
			-- RAID_CLASS_COLORS
			DeathKnight = {0.77, 0.12, 0.23}, -- #C41F3B
			Druid = {1.00 , 0.49, 0.04}, -- #FF7D0A
			Hunter = {0.67, 0.83, 0.45}, -- #ABD473
			Mage = {0.41, 0.80, 0.94}, -- #69CCF0
			Paladin = {0.96, 0.55, 0.73}, -- #F58CBA
			Priest = {1.00, 1.00, 1.00}, -- #FFFFFF
			Rogue = {1.00, 0.96, 0.41}, -- #FFF569
			Shaman = {0.00, 0.44, 0.87}, -- #0070DE
			Warlock = {0.58, 0.51, 0.79}, -- #9482C9
			Warrior = {0.78, 0.61, 0.43}, -- #C79C6E
			-- /dump COMBATLOG_DEFAULT_COLORS.unitColoring[32542]
			Friendly = {0.34, 0.64, 1.00}, -- #57A3FF
			Hostile = {0.75, 0.05, 0.05}, -- #BF0D0D
			Unknown = {191/255, 191/255, 191/255}, -- #BFBFBF
		},
		iconSize = 16,
		iconCropped = true,
		criticalFormat = true,
		filterSelf = true,
		filterEverythingElse = true,
		filterPlayers = true,
		FriendlyDispel = true,
		HostileDispel = true,
		BlizzardCombatLog = true,
		MissTaunt = true,
		MissInterrupt = true,
		MissCC = true,
		sink20OutputSink = "None",
	},
}

local options = {
	type = "group",
	name = "",
	args = {
		Main = {
			type = "group",
			name = " \124cffADFF2FKetho\124r |cffFFFFFFCombatLog|r |cffB6CA00v0.98|r",
			handler = KCL,
			args = {
				groupLocal = {
					type = "group",
					name = function() return profile.ChatFilters and "|cff71D5FFLocal|r" or " " end,
					order = 1,
					inline = true,
					args = {
						LocalTaunt = {
							type = "toggle",
							order = 1,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Reincarnation:20:20:1:0:"..cropped.."|t  |cff"..EventColor["TAUNT"]..GetSpellInfo(355).."|r" end,
							get = function(i) return profile.Taunt end,
							set = function(i, v) profile.Taunt = v end,
						},
						LocalCrowdControl = {
							type = "toggle",
							order = 2,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Polymorph:20:20:1:0:"..cropped.."|t  |cff"..EventColor["CROWDCONTROL"].."Crowd Control|r" end,
							get = function(i) return profile.CrowdControl end,
							set = function(i, v) profile.CrowdControl = v end,
						},
					--	force newline for Blizzard/AceConfigDialog "wide" Options Window; http://forums.wowace.com/showthread.php?p=312303
						newline01 = {type = "description", order = 3, name = ""},
						LocalInterrupt = {
							type = "toggle",
							order = 4,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Kick:20:20:1:0:"..cropped.."|t  |cff"..EventColor["INTERRUPT"]..INTERRUPT.."|r" end,
							get = function(i) return profile.Interrupt end,
							set = function(i, v) profile.Interrupt = v end,
						},
						LocalCC_Break = {
							type = "toggle",
							order = 5,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Seal:20:20:1:0:"..cropped.."|t  |cff"..EventColor["CC_BREAK"].."CC Breaker|r" end,
							get = function(i) return profile.CC_Break end,
							set = function(i, v) profile.CC_Break = v end,
						},
						newline02 = {type = "description", order = 6, name = ""},
						LocalWastedInterrupt = {
							type = "toggle",
							order = 7,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Frost_IceShock:20:20:1:0:"..cropped.."|t  Wasted "..INTERRUPT end,
							disabled = function() return not profile.Interrupt end,
							get = function(i) return profile.WastedInterrupt end,
							set = function(i, v) profile.WastedInterrupt = v end,
						},
						LocalDeath = {
							type = "toggle",
							order = 8,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Rogue_FeignDeath:20:20:1:0:"..cropped.."|t  |cff"..EventColor["DEATH"]..TUTORIAL_TITLE25.."|r" end,
							get = function(i) return profile.Death end,
							set = function(i, v) profile.Death = v end,
						},
						newline03 = {type = "description", order = 9, name = ""},
						LocalDispel = {
							type = "toggle",
							order = 10,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Holy_DispelMagic:20:20:1:0:"..cropped.."|t  |cff"..EventColor["DISPEL"]..GetSpellInfo(25808).."|r" end,
							get = function(i) return profile.Dispel end,
							set = function(i, v) profile.Dispel = v end,
						},
						LocalDeathPrevent = {
							type = "toggle",
							order = 11,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Holy_GuardianSpirit:20:20:1:0:"..cropped.."|t  "..TUTORIAL_TITLE25.." Prevent" end,
							disabled = function() return not profile.Death end,
							get = function(i) return profile.DeathPrevent end,
							set = function(i, v) profile.DeathPrevent = v end,
						},
						newline04 = {type = "description", order = 12, name = ""},
						LocalReflect = {
							type = "toggle",
							order = 13,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Warrior_ShieldReflection:20:20:1:0:"..cropped.."|t  |cff"..EventColor["REFLECT"]..REFLECT.."|r" end,
							get = function(i) return profile.Reflect end,
							set = function(i, v) profile.Reflect = v end,
						},
						LocalResurrection = {
							type = "toggle",
							order = 14,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Holy_Resurrection:20:20:1:0:"..cropped.."|t  |cff"..EventColor["RESURRECTION"]..GetSpellInfo(2006).."|r" end,
							get = function(i) return profile.Resurrection end,
							set = function(i, v) profile.Resurrection = v end,
						},
					},
				},
				groupChat = {
					type = "group",
					name = " ",
					order = 2,
					inline = true,
					name = function() return profile.ChatFilters and "|cff71D5FF"..CHAT.."|r" or " " end,
					hidden = function() return not profile.ChatFilters end,
					args = {
						ChatTaunt = {
							type = "toggle",
							order = 1,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Reincarnation:20:20:1:0:"..cropped.."|t  |cff"..EventColor["TAUNT"]..GetSpellInfo(355).."|r" end,
							get = function(i) return profile.ChatTaunt end,
							set = function(i, v) profile.ChatTaunt = v end,
						},
						ChatCrowdControl = {
							type = "toggle",
							order = 2,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Polymorph:20:20:1:0:"..cropped.."|t  |cff"..EventColor["CROWDCONTROL"].."Crowd Control|r" end,
							get = function(i) return profile.ChatCrowdControl end,
							set = function(i, v) profile.ChatCrowdControl = v end,
						},
					--	force newline for Blizzard/AceConfigDialog "wide" Options Window; http://forums.wowace.com/showthread.php?p=312303
						newline01 = {type = "description", order = 3, name = ""},
						ChatInterrupt = {
							type = "toggle",
							order = 4,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Kick:20:20:1:0:"..cropped.."|t  |cff"..EventColor["INTERRUPT"]..INTERRUPT.."|r" end,
							get = function(i) return profile.ChatInterrupt end,
							set = function(i, v) profile.ChatInterrupt = v end,
						},
						ChatCC_Break = {
							type = "toggle",
							order = 5,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Seal:20:20:1:0:"..cropped.."|t  |cff"..EventColor["CC_BREAK"].."CC Breaker|r" end,
							get = function(i) return profile.ChatCC_Break end,
							set = function(i, v) profile.ChatCC_Break = v end,
						},
						newline02 = {type = "description", order = 6, name = ""},
						ChatWastedInterrupt = {
							type = "toggle",
							order = 7,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Frost_IceShock:20:20:1:0:"..cropped.."|t  Wasted "..INTERRUPT end,
							disabled = function() return not profile.ChatInterrupt end,
							get = function(i) return profile.ChatWastedInterrupt end,
							set = function(i, v) profile.ChatWastedInterrupt = v end,
						},
						ChatDeath = {
							type = "toggle",
							order = 8,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Rogue_FeignDeath:20:20:1:0:"..cropped.."|t  |cff"..EventColor["DEATH"]..TUTORIAL_TITLE25.."|r" end,
							get = function(i) return profile.ChatDeath end,
							set = function(i, v) profile.ChatDeath = v end,
						},
						newline03 = {type = "description", order = 9, name = ""},
						ChatDispel = {
							type = "toggle",
							order = 10,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Holy_DispelMagic:20:20:1:0:"..cropped.."|t  |cff"..EventColor["DISPEL"]..GetSpellInfo(25808).."|r" end,
							get = function(i) return profile.ChatDispel end,
							set = function(i, v) profile.ChatDispel = v end,
						},
						ChatDeathPrevent = {
							type = "toggle",
							order = 11,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Holy_GuardianSpirit:20:20:1:0:"..cropped.."|t  "..TUTORIAL_TITLE25.." Prevent" end,
							disabled = function() return not profile.ChatDeath end,
							get = function(i) return profile.ChatDeathPrevent end,
							set = function(i, v) profile.ChatDeathPrevent = v end,
						},
						newline04 = {type = "description", order = 12, name = ""},
						ChatReflect = {
							type = "toggle",
							order = 13,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Ability_Warrior_ShieldReflection:20:20:1:0:"..cropped.."|t  |cff"..EventColor["REFLECT"]..REFLECT.."|r" end,
							get = function(i) return profile.ChatReflect end,
							set = function(i, v) profile.ChatReflect = v end,
						},
						ChatResurrection = {
							type = "toggle",
							order = 14,
							descStyle = "",
							disabled = "OptionsDisabled",
							name = function() return "|TInterface\\Icons\\Spell_Holy_Resurrection:20:20:1:0:"..cropped.."|t  |cff"..EventColor["RESURRECTION"]..GetSpellInfo(2006).."|r" end,
							get = function(i) return profile.ChatResurrection end,
							set = function(i, v) profile.ChatResurrection = v end,
						},
					},
				},
				spacing = {
					type = "description",
					order = 3,
					name = "",
				},
				togglePvE = {
					type = "toggle",
					order = 4,
					descStyle = "",
					width = "half",
					name = " |cffA8A8FFPvE|r", -- /dump ChatTypeInfo.PARTY {r = 0.666, g = 0.666, b = 1.000}
					disabled = "OptionsDisabled",
					get = function(i) return profile.PvE end,
					set = function(i, v) profile.PvE = v end,
				},
				togglePvP = {
					type = "toggle",
					order = 5,
					width = "half",
					descStyle = "",
					name = " |cffFF7F00"..PVP.."|r", -- /dump ChatTypeInfo.RAID {r = 1.000, g = 0.498, b = 0.000}
					disabled = "OptionsDisabled",
					get = function(i) return profile.PvP end,
					set = function(i, v) profile.PvP = v end,
				},
				toggleWorld = {
					type = "toggle",
					order = 6,
					descStyle = "",
					name = " World",
					width = "half",
					disabled = "OptionsDisabled",
					get = function(i) return profile.World end,
					set = function(i, v) profile.World = v end,
				},
				toggleChatFilters = {
					type = "toggle",
					order = 7,
					descStyle = "",
					disabled = "OptionsDisabled",
					name = "|cff71D5FF"..CHAT.." "..FILTERS.."|r",
					get = function(i) return profile.ChatFilters end,
					set = function(i, v) profile.ChatFilters = v end,
				},
				newline = {
					type = "description",
					order = 8,
					name = "",
				},
				selectChatWindow = {
					type = "select",
					order = 9,
					descStyle = "",
					name = "   |cffFFFFFF"..CHAT.." Window|r",
					disabled = "OptionsDisabled",
					values = function()
						local ChatWindowList = {}
						ChatWindowList[1] = "|cffFF0000<"..ADDON_DISABLED..">|r"
						for i = 1, NUM_CHAT_WINDOWS do
							if GetChatWindowInfo(i) ~= "" then
								ChatWindowList[i+1] = "|cff2E9AFE"..i..".|r  "..select(1, GetChatWindowInfo(i))
							end
						end
						return ChatWindowList end,
					get = function(i) return profile.chatWindow end,
					set = function(i, v)
						profile.chatWindow = v; chatWindow = profile.chatWindow
						if v > 1 then
							ChatFrame = _G["ChatFrame"..v-1]
							ChatFrame:AddMessage("|cffADFF2FKetho CombatLog|r: Chat Frame |cff57A3FF"..(v-1)..".|r |cffADFF2F"..GetChatWindowInfo(v-1).."|r")
						end
					end,
				},
				selectChatChannel = {
					type = "select",
					order = 10,
					descStyle = "",
					name = "   |cffFFFFFF"..CHAT.." "..CHANNEL.."|r",
					disabled = "OptionsDisabled",
					values = function() local ChatChannelList = {}
						ChatChannelList[1] = "|cffFF0000<"..ADDON_DISABLED..">|r"
						ChatChannelList[2] = "|cff2E9AFE#|r   "..CHAT_MSG_SAY
						ChatChannelList[3] = "|cff2E9AFE#|r   |cffFF4040"..CHAT_MSG_YELL.."|r"
						ChatChannelList[4] = "|cff2E9AFE#|r   |cffA8A8FF"..CHAT_MSG_PARTY.."|r / |cffFF7F00"..CHAT_MSG_RAID.."|r"
						for i = 1, 10 do
							local channelID = select((i*2)-1, GetChannelList()) -- this is was kinda hard to code, seeing how GetChannelList() works..
							if channelID then
								ChatChannelList[channelID+4] = "|cff2E9AFE"..channelID..".|r  "..select(i*2,GetChannelList())
							end
						end
						return ChatChannelList end,
					get = function(i) return profile.chatChannel end,
					set = function(i, v) profile.chatChannel = v; chatChannel = profile.chatChannel end,
				},
				descEnable = {
					type = "description",
					order = 11,
					fontSize = "large",
					name = function() return KCL:OptionsDisabled() and " Type |cff2E9AFE/ket on|r to enable" or "" end,
				},
			},
		},
		Advanced = {
			type = "group",
			name = "",
			handler = KCL,
			args = {
				groupColoring = {
					type = "group",
					order = 1,
					name = "|TInterface\\Icons\\INV_Misc_Gem_Variety_02:14:14:1:0:"..cropped.."|t  |cffFFFFFF"..COLORS.."|r",
					args = {
						colorTaunt = {
							type = "color",
							order = 1,
							name = "|TInterface\\Icons\\Spell_Nature_Reincarnation:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(355),
							get = function(i) return unpack(color.Taunt) end,
							set = function(info, r,g,b) color.Taunt = {r, g, b}; EventColor["TAUNT"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorInterrupt = {
							type = "color",
							order = 3,
							name = "|TInterface\\Icons\\Ability_Kick:16:16:1:0:"..cropped.."|t  "..INTERRUPT,
							get = function(i) return unpack(color.Interrupt) end,
							set = function(info, r,g,b) color.Interrupt = {r, g, b}; EventColor["INTERRUPT"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorDispel = {
							type = "color",
							order = 5,
							name = "|TInterface\\Icons\\Spell_Holy_DispelMagic:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(25808),
							get = function(i) return unpack(color.Dispel) end,
							set = function(info, r,g,b) color.Dispel = {r, g, b}; EventColor["DISPEL"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorReflect = {
							type = "color",
							order = 7,
							name = "|TInterface\\Icons\\Ability_Warrior_ShieldReflection:16:16:1:0:"..cropped.."|t  "..REFLECT,
							get = function(i) return unpack(color.Reflect) end,
							set = function(info, r,g,b) color.Reflect = {r, g, b}; EventColor["REFLECT"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorCrowdControl = {
							type = "color",
							order = 2,
							name = "|TInterface\\Icons\\Spell_Nature_Polymorph:16:16:1:0:"..cropped.."|t  Crowd Control",
							get = function(i) return unpack(color.CrowdControl) end,
							set = function(info, r,g,b) color.CrowdControl = {r, g, b}; EventColor["CROWDCONTROL"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorCC_Break = {
							type = "color",
							order = 4,
							name = "|TInterface\\Icons\\Ability_Seal:16:16:1:0:"..cropped.."|t  CC Breaker",
							get = function(i) return unpack(color.CC_Break) end,
							set = function(info, r,g,b) color.CC_Break = {r, g, b}; EventColor["CC_BREAK"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorDeath = {
							type = "color",
							order = 6,
							name = "|TInterface\\Icons\\Ability_Rogue_FeignDeath:16:16:1:0:"..cropped.."|t  "..TUTORIAL_TITLE25,
							get = function(i) return unpack(color.Death) end,
							set = function(info, r,g,b) color.Death = {r, g, b}; EventColor["DEATH"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorResurrection = {
							type = "color",
							order = 8,
							name = "|TInterface\\Icons\\Spell_Holy_Resurrection:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(2006),
							get = function(i) return unpack(color.Resurrection) end,
							set = function(info, r,g,b) color.Resurrection = {r, g, b}; EventColor["RESURRECTION"] = KCL:Dec2Hex(r, g, b) end,
						},
						header01 = {type = "header", order = 9, name = ""},
						colorWarrior = {
							type = "color",
							order = 10,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:4:60:4:60|t  "..LOCALIZED_CLASS_NAMES_MALE["WARRIOR"],
							get = function(i) return unpack(color.Warrior) end,
							set = function(info, r,g,b) color.Warrior = {r, g, b}; ClassColor["WARRIOR"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorDeathKnight = {
							type = "color",
							order = 12,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:68:124:133:189|t  "..LOCALIZED_CLASS_NAMES_MALE["DEATHKNIGHT"],
							get = function(i) return unpack(color.DeathKnight) end,
							set = function(info, r,g,b) color.DeathKnight = {r, g, b}; ClassColor["DEATHKNIGHT"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorPaladin = {
							type = "color",
							order = 14,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:4:60:132:188|t  "..LOCALIZED_CLASS_NAMES_MALE["PALADIN"],
							get = function(i) return unpack(color.Paladin) end,
							set = function(info, r,g,b) color.Paladin = {r, g, b}; ClassColor["PALADIN"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorPriest = {
							type = "color",
							order = 16,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:130:186:68:124|t  "..LOCALIZED_CLASS_NAMES_MALE["PRIEST"],
							get = function(i) return unpack(color.Priest) end,
							set = function(info, r,g,b) color.Priest = {r, g, b}; ClassColor["PRIEST"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorShaman = {
							type = "color",
							order = 18,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:68:124:68:124|t  "..LOCALIZED_CLASS_NAMES_MALE["SHAMAN"],
							get = function(i) return unpack(color.Shaman) end,
							set = function(info, r,g,b) color.Shaman = {r, g, b}; ClassColor["SHAMAN"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorDruid = {
							type = "color",
							order = 11,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:194:250:4:60|t  "..LOCALIZED_CLASS_NAMES_MALE["DRUID"],
							get = function(i) return unpack(color.Druid) end,
							set = function(info, r,g,b) color.Druid = {r, g, b}; ClassColor["DRUID"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorRogue = {
							type = "color",
							order = 13,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:131:187:4:60|t  "..LOCALIZED_CLASS_NAMES_MALE["ROGUE"],
							get = function(i) return unpack(color.Rogue) end,
							set = function(info, r,g,b) color.Rogue = {r, g, b}; ClassColor["ROGUE"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorMage = {
							type = "color",
							order = 15,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:68:124:4:60|t  "..LOCALIZED_CLASS_NAMES_MALE["MAGE"],
							get = function(i) return unpack(color.Mage) end,
							set = function(info, r,g,b) color.Mage = {r, g, b}; ClassColor["MAGE"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorWarlock = {
							type = "color",
							order = 17,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:194:250:68:124|t  "..LOCALIZED_CLASS_NAMES_MALE["WARLOCK"],
							get = function(i) return unpack(color.Warlock) end,
							set = function(info, r,g,b) color.Warlock = {r, g, b}; ClassColor["WARLOCK"] = KCL:Dec2Hex(r, g, b) end,
						},
						colorHunter = {
							type = "color",
							order = 19,
							name = "|TInterface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES:16:16:1:0:256:256:4:60:68:124|t  "..LOCALIZED_CLASS_NAMES_MALE["HUNTER"],
							get = function(i) return unpack(color.Hunter) end,
							set = function(info, r,g,b) color.Hunter = {r, g, b}; ClassColor["HUNTER"] = KCL:Dec2Hex(r, g, b) end,
						},
						header02 = {type = "header", order = 20, name = ""},
						colorPhysical = {
							type = "color",
							order = 21,
							name = "|TInterface\\Icons\\Spell_Nature_Strength:16:16:1:0:"..cropped.."|t  "..STRING_SCHOOL_PHYSICAL,
							get = function(i) return unpack(color.Physical) end,
							set = function(info, r,g,b) color.Physical = {r, g, b}; SpellSchoolColor[0x1] = KCL:Dec2Hex(r, g, b) end,
						},
						colorHoly = {
							type = "color",
							order = 23,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon2:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_HOLY,
							get = function(i) return unpack(color.Holy) end,
							set = function(info, r,g,b) color.Holy = {r, g, b}; SpellSchoolColor[0x2] = KCL:Dec2Hex(r, g, b) end,
						},
						colorFire = {
							type = "color",
							order = 25,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon3:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_FIRE,
							get = function(i) return unpack(color.Fire) end,
							set = function(info, r,g,b) color.Fire = {r, g, b}; SpellSchoolColor[0x4] = KCL:Dec2Hex(r, g, b) end,
						},
						colorNature = {
							type = "color",
							order = 27,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon4:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_NATURE,
							get = function(i) return unpack(color.Nature) end,
							set = function(info, r,g,b) color.Nature = {r, g, b}; SpellSchoolColor[0x8] = KCL:Dec2Hex(r, g, b) end,
						},
						colorFrost = {
							type = "color",
							order = 22,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon5:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_FROST,
							get = function(i) return unpack(color.Frost) end,
							set = function(info, r,g,b) color.Frost = {r, g, b}; SpellSchoolColor[0x10] = KCL:Dec2Hex(r, g, b) end,
						},
						colorShadow = {
							type = "color",
							order = 24,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon6:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_SHADOW,
							get = function(i) return unpack(color.Shadow) end,
							set = function(info, r,g,b) color.Shadow = {r, g, b}; SpellSchoolColor[0x20] = KCL:Dec2Hex(r, g, b) end,
						},
						colorArcane = {
							type = "color",
							order = 26,
							name = "|TInterface\\PaperDollInfoFrame\\SpellSchoolIcon7:16:16:1:0:16:16:1:15:1:15|t  "..STRING_SCHOOL_ARCANE,
							get = function(i) return unpack(color.Arcane) end,
							set = function(info, r,g,b) color.Arcane = {r, g, b}; SpellSchoolColor[0x40] = KCL:Dec2Hex(r, g, b) end,
						},
						header03 = {type = "header", order = 28, name = ""},
						colorFriendly = {
							type = "color",
							order = 29,
							width = "full",
							name = "|TInterface\\Icons\\Spell_ChargePositive:16:16:1:0:"..cropped.."|t  "..FRIENDLY,
							get = function(i) return unpack(color.Friendly) end,
							set = function(info, r,g,b) color.Friendly = {r, g, b}; UnitColor[1] = KCL:Dec2Hex(r, g, b) end,
						},
						colorHostile = {
							type = "color",
							order = 30,
							width = "full",
							name = "|TInterface\\Icons\\Spell_ChargeNegative:16:16:1:0:"..cropped.."|t  "..HOSTILE,
							get = function(i) return unpack(color.Hostile) end,
							set = function(info, r,g,b) color.Hostile = {r, g, b}; UnitColor[2] = KCL:Dec2Hex(r, g, b) end,
						},
						colorUnknown = {
							type = "color",
							order = 31,
							width = "full",
							name = "|TInterface\\Icons\\INV_Misc_QuestionMark:16:16:1:0:"..cropped.."|t  "..UNKNOWN,
							get = function(i) return unpack(color.Unknown) end,
							set = function(info, r,g,b) color.Unknown = {r, g, b}; UnitColor[3] = KCL:Dec2Hex(r, g, b) end,
						},
						spacing = {type = "description", order = 32, name = ""},
						executeReset = {
							type = "execute",
							order = 33,
							descStyle = "",
							name = "|cffFFFFFFReset Colors|r",
							confirm = true, confirmText = "Reset Colors?",
							-- for some reason the colors that are "called/copied" from also get changed,
							-- so the reset function only works once?!
						--	func = function() profile.color = defaults.profile.color; KCL:RefreshConfig() end,
							func = function()
								profile.color = {
									Taunt = {1, 0, 0},
									Interrupt = {0, 110/255, 1},
									Dispel = {1, 1, 1},
									Reflect = {1, 1, 1},
									CrowdControl = {1, 1, 1},
									CC_Break = {1, 1, 1},
									Death = {1, 1, 1},
									Resurrection = {175/255, 1, 47/255},
									Physical = {1.00, 1.00, 0.00},
									Holy = {1.00, 0.90, 0.50},
									Fire = {1.00, 0.50, 0.00},
									Nature = {0.30, 1.00, 0.30},
									Frost = {0.50, 1.00, 1.00},
									Shadow = {0.50, 0.50, 1.00},
									Arcane = {1.00, 0.50, 1.00},
									DeathKnight = {0.77, 0.12, 0.23},
									Druid = {1.00 , 0.49, 0.04},
									Hunter = {0.67, 0.83, 0.45},
									Mage = {0.41, 0.80, 0.94},
									Paladin = {0.96, 0.55, 0.73},
									Priest = {1.00, 1.00, 1.00},
									Rogue = {1.00, 0.96, 0.41},
									Shaman = {0.00, 0.44, 0.87},
									Warlock = {0.58, 0.51, 0.79},
									Warrior = {0.78, 0.61, 0.43},
									Friendly = {0.34, 0.64, 1.00},
									Hostile = {0.75, 0.05, 0.05},
									Unknown = {191/255, 191/255, 191/255},
								}
								KCL:RefreshConfig()
							end,
						},
					},
				},
				groupIcon = {
					type = "group",
					order = 2,
					name = "|TINTERFACE\\ICONS\\inv_misc_orb_05:14:14:1:0:"..cropped.."|t  |cffFFFFFFIcons|r",
					args = {
						selectIconSize = {
							type = "select",
							order = 1,
							descStyle = "", 
							name = "|cffFFFFFF"..EMBLEM_SYMBOL.." Size|r",
							values = {"|cffFF0000<"..ADDON_DISABLED..">|r", nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, " 12", nil, " 14", nil, " 16  |cffFBDB00(Default)|r", nil, " 18",  nil, " 20", nil, " 22", nil, " 24", nil, " 26", nil, " 28", nil, " 30", nil, " 32"},
							get = function(i) return profile.iconSize end,
							set = function(i, v) profile.iconSize = v; iconSize = profile.iconSize end,
						},
						newlin = {type = "description", order = 2, name = " "},
						toggleCropIcon = {
							type = "toggle",
							order = 3,
							descStyle = "",
							name = "|TInterface\\Icons\\inv_misc_orb_05:20:20:1:0|t |TInterface\\Icons\\inv_misc_orb_05:20:20:1:0:64:64:4:60:4:60|t  Crop",
							disabled = function() return not profile.iconSize end,
							get = function(i) return profile.iconCropped end,
							set = function(i, v)
								profile.iconCropped = v
								cropped = v and "64:64:4:60:4:60" or ""
							end,
						},
					}
				},
				groupOther = {
					type = "group",
					order = 3,
					name = "|TInterface\\Icons\\Trade_Engineering:14:14:1:0:"..cropped.."|t  |cffFFFFFF"..OTHER.."|r",
					args = {
						header1 = {type = "header", order = 1, name = "Message Format"},
						toggleOverkillFormat = {
							type = "toggle",
							order = 2,
							desc = "<message> |cff71D5FF"..TEXT_MODE_A_STRING_RESULT_OVERKILLING.."|r",
							name = OVERKILLING,
							get = function(i) return profile.overkillFormat end,
							set = function(i, v) profile.overkillFormat = v; overkillFormat = profile.overkillFormat end,
						},
						toggleCriticalFormat = {
							type = "toggle",
							order = 4,
							desc = "<message> |cff71D5FF"..TEXT_MODE_A_STRING_RESULT_CRITICAL.."|r",
							name = CRITICAL,
							get = function(i) return profile.criticalFormat end,
							set = function(i, v) profile.criticalFormat = v; criticalFormat = profile.criticalFormat end,
						},
						toggleGlancingFormat = {
							type = "toggle",
							order = 3,
							desc = "<message> |cff71D5FF"..TEXT_MODE_A_STRING_RESULT_GLANCING .."|r",
							name = GLANCING,
							get = function(i) return profile.glancingFormat end,
							set = function(i, v) profile.glancingFormat = v; glancingFormat = profile.glancingFormat end,
						},
						toggleCrushingFormat = {
							type = "toggle",
							order = 5,
							desc = "<message> |cff71D5FF"..TEXT_MODE_A_STRING_RESULT_CRUSHING.."|r",
							name = CRUSHING,
							get = function(i) return profile.crushingFormat end,
							set = function(i, v) profile.crushingFormat = v; crushingFormat = profile.crushingFormat end,
						},
						toggleTimestamp = {
							type = "toggle",
							order = 6,
							desc = TIMESTAMP_COMBATLOG_TOOLTIP,
							name = TIMESTAMPS_LABEL,
							get = function(i) return profile.Timestamp end,
							set = function(i, v) profile.Timestamp = v; Timestamp = profile.Timestamp end,
						},
						header3 = {type = "header", order = 7, name = OTHER},
						toggleBlizzardCombatLog = {
							type = "toggle",
							order = 8,
							desc = BINDING_NAME_TOGGLECOMBATLOG,
							width = "full",
							name = "Blizzard CombatLog",
							get = function(i) return profile.BlizzardCombatLog end,
							set = function(i, v) profile.BlizzardCombatLog = v
								if v then
									COMBATLOG:RegisterEvent("COMBAT_LOG_EVENT")
									print("Blizzard CombatLog: |cffB6CA00Enabled|r")
								else
									COMBATLOG:UnregisterEvent("COMBAT_LOG_EVENT")
									print("Blizzard CombatLog: |cffFF2424Disabled|r")
								end
							end,
						},
						toggleLoggingChat = {
							type = "toggle",
							order = 9,
							desc = "Note: |cff71D5FFPrat|r can override these settings",
							width = "full",
							name = CHAT.." Logging",
							get = function() return LoggingChat() end,
							set = SlashCmdList["CHATLOG"],
						},
						toggleLoggingCombat = {
							type = "toggle",
							order = 10,
							descStyle = "",
							width = "full",
							name = COMBAT.." Logging",
							get = function() return LoggingCombat() end,
							set = SlashCmdList["COMBATLOG"],
						},
					},
				},
				groupSpacing1 = {
					type = "group",
					order = 4,
					name = "",
					disabled = true,
					args = {},
				},
				groupFilter = {
					type = "group",
					order = 5,
					name = "|TInterface\\Icons\\Spell_Holy_Silence:14:14:1:0:"..cropped.."|t  |cffFFFFFF"..FILTERS.."|r",
					args = {
						header1 = {type = "header", order = 1, name = COMBAT_LOG_MENU_BOTH},
						toggleSelf = {
							type = "toggle",
							order = 2,
							descStyle = "",
							width = "full",
							name = YOU,
							get = function(i) return profile.filterSelf end,
							set = function(i, v) profile.filterSelf = v
								filterSelf = profile.filterSelf
								if not v then
									filterEverythingElse, profile.filterEverythingElse = true, true
								end
							end,
						},
						toggleEverythingElse = {
							type = "toggle",
							order = 3,
							descStyle = "",
							name = "Others",
							get = function(i) return profile.filterEverythingElse end,
							set = function(i, v) profile.filterEverythingElse = v 
								filterEverythingElse = profile.filterEverythingElse
								if not v then
									filterSelf, profile.filterSelf = true, true
								end
							end,
						},
						header2 = {type = "header", order = 4, name = BY_SOURCE.." "..TYPE},
						togglePlayers = {
							type = "toggle",
							order = 5,
							descStyle = "",
							width = "full",
							name = TUTORIAL_TITLE19,
							get = function(i) return profile.filterPlayers end,
							set = function(i, v) profile.filterPlayers = v
								filterPlayers = profile.filterPlayers
								if not v then
									filterMonsters, profile.filterMonsters = true, true
								end
							end,
						},
						toggleMonsters = {
							type = "toggle",
							order = 5,
							descStyle = "",
							width = "full",
							name = "NPCs",
							get = function(i) return profile.filterMonsters end,
							set = function(i, v) profile.filterMonsters = v 
								filterMonsters = profile.filterMonsters
								if not v then
									filterPlayers, profile.filterPlayers = true, true
								end
							end,
						},
						header3 = {type = "header", order = 6, name = TANK},
						toggleTankTaunt = {
							type = "toggle",
							order = 7,
							descStyle = "",
							width = "full",
							name = function() return "|TInterface\\LFGFRAME\\UI-LFG-ICON-ROLES:20:20:-3:0:256:256:1:65:68:132|t|TInterface\\Icons\\Spell_Nature_Reincarnation:16:16:0:0:"..cropped.."|t  |cff6FAFE6"..TANK.."|r Taunt" end,
							disabled = function() return not profile.Taunt end,
							get = function(i) return profile.TankTaunt end,
							set = function(i, v) profile.TankTaunt = v end,
						},
						toggleTankBreak = {
							type = "toggle",
							order = 8,
							descStyle = "",
							width = "full",
							name = function() return "|TInterface\\LFGFRAME\\UI-LFG-ICON-ROLES:20:20:-3:0:256:256:1:65:68:132|t|TInterface\\Icons\\Ability_Seal:16:16:0:0:"..cropped.."|t  |cff6FAFE6"..TANK.."|r CC Break" end,
							disabled = function() return not profile.CC_Break end,
							get = function(i) return profile.TankBreak end,
							set = function(i, v) profile.TankBreak = v end,
						},
						togglePetGrowl = {
							type = "toggle",
							order = 9,
							descStyle = "",
							width = "full",
							name = function() return " |TInterface\\Icons\\INV_Weapon_Bow_07:16:16:-4:0:"..cropped.."|t|TInterface\\Icons\\Ability_Physical_Taunt:16:16:1:0:"..cropped.."|t  |cffABD473"..PET.."|r "..GetSpellInfo(2649) end,
							disabled = function() return not profile.Taunt end,
							get = function(i) return profile.PetGrowl end,
							set = function(i, v) profile.PetGrowl = v end,
						},
						header4 = {type = "header", order = 10, name = "|TInterface\\Icons\\Spell_Holy_DispelMagic:16:16:1:0:"..cropped.."|t  "..DISPELS},
						toggleFriendlyDispel = {
							type = "toggle",
							order = 11,
							descStyle = "",
							name = FRIENDLY.." (|cff71D5FF"..ACTION_SPELL_DISPEL_DEBUFF.."|r)",
							disabled = function() return not profile.Dispel end,
							get = function(i) return profile.FriendlyDispel end,
							set = function(i, v) profile.FriendlyDispel = v end,
						},
						toggleHostileDispel = {
							type = "toggle",
							order = 12,
							descStyle = "",
							width = "full",
							name = HOSTILE.." (|cff71D5FF"..ACTION_SPELL_DISPEL_BUFF.."|r)",
							disabled = function() return not profile.Dispel end,
							get = function(i) return profile.HostileDispel end,
							set = function(i, v) profile.HostileDispel = v end,
						},
					},
				},
				groupEvent = {
					type = "group",
					order = 6,
					name = "|TInterface\\Icons\\Ability_Marksmanship:14:14:1:0:"..cropped.."|t  |cffFFFFFFExtra Events|r",
					args = {
						header1 = {type = "header", order = 1, name = "|TInterface\\Icons\\Ability_Marksmanship:14:14:1:0:"..cropped.."|t "..MISSES},
						toggleMissTaunt = {
							type = "toggle",
							order = 2,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Reincarnation:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(355) end,
							disabled = function() return not profile.Taunt or profile.MissAll end,
							get = function(i) return profile.MissTaunt end,
							set = function(i, v) profile.MissTaunt = v end,
						},
						newline01 = {type = "description", order = 3, name = ""},
						toggleMissInterrupt = {
							type = "toggle",
							order = 4,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Ability_Kick:16:16:1:0:"..cropped.."|t  "..INTERRUPT end,
							disabled = function() return not profile.Interrupt or profile.MissAll end,
							get = function(i) return profile.MissInterrupt end,
							set = function(i, v) profile.MissInterrupt = v end,
						},
						newline02 = {type = "description", order = 5, name = ""},
						toggleMissCC = {
							type = "toggle",
							order = 6,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_Nature_Polymorph:16:16:1:0:"..cropped.."|t  Crowd Control" end,
							disabled = function() return not profile.CrowdControl or profile.MissAll end,
							get = function(i) return profile.MissCC end,
							set = function(i, v) profile.MissCC = v end,
						},
						newline03 = {type = "description", order = 7, name = ""},
						toggleMissAll = {
							type = "toggle",
							order = 8,
							descStyle = "",
							name = function() return "|TInterface\\Icons\\Spell_ChargePositive:16:16:1:0:"..cropped.."|t  |cff71D5FF"..ALL.."|r" end,
							disabled = "OptionsDisabled",
							get = function(i) return profile.MissAll end,
							set = function(i, v) profile.MissAll = v end,
						},						
					},
				},
				groupSpacing = {
					type = "group",
					order = 7,
					name = "",
					disabled = true,
					args = {}
				},
				groupAbility = {
					type = "group",
					order = 8,
					disabled = true,
					name = "|TInterface\\Icons\\Ability_Rogue_TricksOftheTrade:14:14:1:0:"..cropped.."|t  "..COMBATLOG_HIGHLIGHT_ABILITY.."  |cffFF0000(NYI)|r",
					args = {
						toggleSoulstoneResurrection = {
							type = "toggle",
							order = 1,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Shadow_SoulGem:16:16:1:0:"..cropped.."|t  |cff9482C9"..GetSpellInfo(20707).."|r",
							get = function(i) return profile.abilitySoulstone end,
							set = function(i, v) profile.abilitySoulstone = v end,
						},
						toggleFocusMagic = {
							type = "toggle",
							order = 2,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Arcane_StudentOfMagic:16:16:1:0:"..cropped.."|t  |cff69CCF0"..GetSpellInfo(54646).."|r",
							get = function(i) return profile.abilityFocusMagic end,
							set = function(i, v) profile.abilityFocusMagic = v end,
						},
						toggleVigilance = {
							type = "toggle",
							order = 3,
							desc = "", 
							name = "|TInterface\\Icons\\Ability_Warrior_Vigilance:16:16:1:0:"..cropped.."|t  |cffC79C6E"..GetSpellInfo(50720).."|r",
							get = function(i) return profile.abilityVigilance end,
							set = function(i, v) profile.abilityVigilance = v end,
						},
						toggleMisdirection = {
							type = "toggle",
							order = 4,
							desc = "", 
							name = "|TInterface\\Icons\\Ability_Hunter_Misdirection:16:16:1:0:"..cropped.."|t  |cffABD473"..GetSpellInfo(34477).."|r",
							get = function(i) return profile.abilityMisdirection end,
							set = function(i, v) profile.abilityMisdirection = v end,
						},
						toggleTricksTrade = {
							type = "toggle",
							order = 5,
							desc = "", 
							name = "|TInterface\\Icons\\Ability_Rogue_TricksOftheTrade:16:16:1:0:"..cropped.."|t  |cffFFF569"..GetSpellInfo(57934).."|r",
							get = function(i) return profile.abilityTricksTrade end,
							set = function(i, v) profile.abilityTricksTrade = v end,
						},
						toggleInnervate = {
							type = "toggle",
							order = 6,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Nature_Lightning:16:16:1:0:"..cropped.."|t  |cffFF7D0A"..GetSpellInfo(29166).."|r",
							get = function(i) return profile.abilityInnervate end,
							set = function(i, v) profile.abilityInnervate = v end,
						},
						toggleHeroismBloodlust = {
							type = "toggle",
							order = 7,
							desc = "", 
							name = "|TInterface\\Icons\\Ability_Shaman_Heroism:16:16:1:0:"..cropped.."|t |TInterface\\Icons\\Spell_Nature_BloodLust:16:16:1:0:"..cropped.."|t  |cff0070DE"..GetSpellInfo(32182).."|r / |cff0070DE"..GetSpellInfo(2825).."|r",
							get = function(i) return profile.abilityBloodlust end,
							set = function(i, v) profile.abilityBloodlust = v end,
						},
						toggleRitualRefreshment = {
							type = "toggle",
							order = 8,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Arcane_MassDispel:16:16:1:0:"..cropped.."|t  |cff69CCF0"..GetSpellInfo(43987).."|r",
							get = function(i) return profile.abilityRitualRefreshment end,
							set = function(i, v) profile.abilityRitualRefreshment = v end,
						},
						toggleFearWard = {
							type = "toggle",
							order = 9,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Holy_Excorcism:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(6346).."",
							get = function(i) return profile.abilityFearWard end,
							set = function(i, v) profile.abilityFearWard = v end,
						},
						toggleMindVision = {
							type = "toggle",
							order = 10,
							desc = "", 
							name = "|TInterface\\Icons\\Spell_Holy_MindVision:16:16:1:0:"..cropped.."|t  "..GetSpellInfo(2096).."",
							get = function(i) return profile.abilityMindVision end,
							set = function(i, v) profile.abilityMindVision = v end,
						},
					},
				},
				groupFun = {
					type = "group",
					order = 9,
					disabled = true,
					name = "|TInterface\\AddOns\\Ketho_CombatLog\\Images\\Awesome:14:14:1:0|t  Fun  |cffFF0000(NYI)|r",
					args = {
						toggleFeast = {
							type = "toggle",
							order = 1,
							desc = "|cffF6ADC6\"om nom nom nom\"|r", 
							name = "  Feasts",
							get = function(i) return profile.Feast end,
							set = function(i, v) profile.Feast = v end,
						},
						descriptionFeasts = {
							type = "description",
							order = 2,
							name = "   |TInterface\\Icons\\Spell_Misc_Food:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Food_Meat_Cooked_02:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Food_99:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Fish_52:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\Ability_Hunter_Pet_Boar:16:16:0:0:"..cropped.."|t",
						},
						header2 = {type = "header", order = 3, name = ""},
						toggleRepairBot = {
							type = "toggle",
							order = 4,
							desc = "", 
							name = "  Repair Bots",
							get = function(i) return profile.RepairBot end,
							set = function(i, v) profile.RepairBot = v end,
						},
						descriptionRepairBot = {
							type = "description",
							order = 5,
							name = "   |TInterface\\Minimap\\Tracking\\Repair:16:16:-2:0:"..cropped.."|t |TInterface\\Icons\\inv_pet_lilsmoky:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_EngGizmos_14:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\inv_misc_molle:16:16:0:0:"..cropped.."|t",
						},
						header3 = {type = "header", order = 6, name = ""},
						toggleFun = {
							type = "toggle",
							order = 7,
							desc = "|cffF6ADC6\"trolololol\"|r", 
							name = "  Fun",
							get = function(i) return profile.Fun end,
							set = function(i, v) profile.Fun = v end,
						},
						descriptionFun = {
							type = "description",
							order = 8,
							name = "   |TInterface\\Icons\\inv_misc_mohawkgrenade:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Toy_10:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Bag_28_Halloween:16:16:0:0:"..cropped.."|t |TInterface\\Icons\\INV_Misc_Sandbox_Spectraltiger_01:16:16:0:0:"..cropped.."|t",
						},
					},
				},
			},
		},
	},
}

	-----------------------------------
	---- 03  OnInitialize, OnEnable ---
	-----------------------------------

function KCL:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("KethoCombatLogDB", defaults, true)

	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	self:RefreshConfig()

	self.db.global.version = 0.98
	self.db.global.fileType = "Release"

	ACR:RegisterOptionsTable("KethoCombatLog_Main", options.args.Main)
	ACR:RegisterOptionsTable("KethoCombatLog_Advanced", options.args.Advanced)
	ACR:RegisterOptionsTable("KethoCombatLog_LibSink", self:GetSinkAce3OptionsDataTable())
	ACR:RegisterOptionsTable("KethoCombatLog_Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db))

	local Main = "Ketho |cffFFFFFFCombatLog|r"
	optionsFrame = ACD:AddToBlizOptions("KethoCombatLog_Main", Main)
	ACD:AddToBlizOptions("KethoCombatLog_Advanced", "|TInterface\\Icons\\Trade_Engineering:16:16:1:0:"..cropped.."|t  "..ADVANCED, Main)
	ACD:AddToBlizOptions("KethoCombatLog_LibSink", "|TINTERFACE\\ICONS\\inv_scroll_03:16:16:1:0:"..cropped.."|t  LibSink", Main)
	ACD:AddToBlizOptions("KethoCombatLog_Profiles", "|TInterface\\Icons\\INV_Misc_Note_01:16:16:1:0:"..cropped.."|t  Profiles", Main)

	self:RegisterChatCommand("kcl", "SlashCommand")
	self:RegisterChatCommand("ket", "SlashCommand")
	self:RegisterChatCommand("ketho", "SlashCommand")
	self:RegisterChatCommand("kethocombat", "SlashCommand")
	self:RegisterChatCommand("kethocombatlog", "SlashCommand")

	if not profile.enableAddOn then
		-- Turn AddOn off again
		self:Disable()
	end
end

local instanceType, instanceTypeFilter
local chatType, channel

function KCL:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("ADDON_LOADED") -- Check for Blizzard CombatLog

	self:ScheduleRepeatingTimer(function()
		-- zone based, instead of group based "detection"
		instanceType = select(2, IsInInstance())

		if chatChannel == 2 then
			chatType = "SAY"
		elseif chatChannel == 3 then
			chatType = "YELL"
		elseif chatChannel == 4 then
			chatType = instanceType == "raid" and "RAID" or (instanceType == "party" or instanceType == "arena" or GetNumPartyMembers() > 0) and "PARTY"
		else
			chatType, channel = "CHANNEL", chatChannel-4
		end

		if (profile.PvE and (instanceType == "party" or instanceType == "raid")) or
		(profile.PvP and (instanceType == "pvp" or instanceType == "arena")) or
		(profile.World and instanceType == "none") then
			instanceTypeFilter = true
		else
			instanceTypeFilter = false
		end
	end, 3)
end

	------------------------------
	--- 04  Callback Functions ---
	------------------------------

function KCL:RefreshConfig()
	-- table reference lookup shortcuts
	profile, char = self.db.profile, self.db.char
	color = profile.color

	-- LibSink
	self:SetSinkStorage(self.db.profile)

	-- optimize for CLEU
	filterSelf, filterEverythingElse = profile.filterSelf, profile.filterEverythingElse
	filterPlayers, filterMonsters = profile.filterPlayers, profile.filterMonsters
	chatWindow, chatChannel = profile.chatWindow, profile.chatChannel
	iconSize, Timestamp = profile.iconSize, profile.Timestamp
	overkillFormat, criticalFormat = profile.overkillFormat, profile.criticalFormat
	glancingFormat, crushingFormat = profile.glancingFormat, profile.crushingFormat

	-- other
	cropped = profile.iconCropped and "64:64:4:60:4:60" or ""
	if chatWindow > 1 then
		ChatFrame = _G["ChatFrame"..chatWindow-1]
	end

	-- update color arrays
	ClassColor = {
		["DEATHKNIGHT"] = self:Dec2Hex(unpack(color.DeathKnight)),
		["DRUID"] = self:Dec2Hex(unpack(color.Druid)),
		["HUNTER"] = self:Dec2Hex(unpack(color.Hunter)),
		["MAGE"] = self:Dec2Hex(unpack(color.Mage)),
		["PALADIN"] = self:Dec2Hex(unpack(color.Paladin)),
		["PRIEST"] = self:Dec2Hex(unpack(color.Priest)),
		["ROGUE"] = self:Dec2Hex(unpack(color.Rogue)),
		["SHAMAN"] = self:Dec2Hex(unpack(color.Shaman)),
		["WARLOCK"] = self:Dec2Hex(unpack(color.Warlock)),
		["WARRIOR"] = self:Dec2Hex(unpack(color.Warrior)),
	}

	SpellSchoolColor = {
		[0x1] = self:Dec2Hex(unpack(color.Physical)),
		[0x2] = self:Dec2Hex(unpack(color.Holy)),
		[0x4] = self:Dec2Hex(unpack(color.Fire)),
		[0x8] = self:Dec2Hex(unpack(color.Nature)),
		[0x10] = self:Dec2Hex(unpack(color.Frost)),
		[0x20] = self:Dec2Hex(unpack(color.Shadow)),
		[0x40] = self:Dec2Hex(unpack(color.Arcane)),
	}

	UnitColor = {
		self:Dec2Hex(unpack(color.Friendly)),
		self:Dec2Hex(unpack(color.Hostile)),
		self:Dec2Hex(unpack(color.Unknown)),
	}
	
	EventColor = {
		["TAUNT"] = self:Dec2Hex(unpack(color.Taunt)),
		["INTERRUPT"] = self:Dec2Hex(unpack(color.Interrupt)),
		["DISPEL"] = self:Dec2Hex(unpack(color.Dispel)),
		["REFLECT"] = self:Dec2Hex(unpack(color.Reflect)),
		["CROWDCONTROL"] = self:Dec2Hex(unpack(color.CrowdControl)),
		["CC_BREAK"] = self:Dec2Hex(unpack(color.CC_Break)),
		["DEATH"] = self:Dec2Hex(unpack(color.Death)),
		["RESURRECTION"] = self:Dec2Hex(unpack(color.Resurrection)),
	}
end

function KCL:OptionsDisabled()
	return not self:IsEnabled()
end

function KCL:SlashCommand(input)
	if #strtrim(input) == 0 then
		InterfaceOptionsFrame_OpenToCategory(optionsFrame)

	elseif input == "enable" or input == "on" or input == "1" then
		if not profile.enableAddOn then
			self:Enable()
			profile.enableAddOn = true
			self:Print("|cffADFF2FEnabled|r")
		else
			self:Print("Already |cffADFF2FEnabled|r")
		end
		ACR:NotifyChange("KethoCombatLog_Main")
	elseif input == "disable" or input == "off" or input == "0" then
		if profile.enableAddOn then
			self:Disable()
			profile.enableAddOn = false
			self:Print("|cffFF2424Disabled|r")
		else
			self:Print("Already |cffFF2424Disabled|r")
		end
		ACR:NotifyChange("KethoCombatLog_Main")
	else
		print("|cff2E9AFE/ket|r: Open Options\n|cff2E9AFE/ket|r |cffB6CA00on|r: Enable AddOn\n|cff2E9AFE/ket|r |cffFF2424off|r: Disable AddOn")
	end
end

-- Blizzard CombatLog is Load-on-Demand
function KCL:ADDON_LOADED(event, addon)
	if addon == "Blizzard_CombatLog" and not profile.BlizzardCombatLog then
		COMBATLOG:UnregisterEvent("COMBAT_LOG_EVENT")
		self:UnregisterEvent("ADDON_LOADED")
	end
end

local function ChatFilter(subevent)
	if profile.ChatFilters then
		return profile["Chat"..subevent]
	else
		return profile[subevent]
	end
end

function KCL:Dec2Hex(R, G, B)
	return format("%02X%02X%02X", R*255, G*255, B*255)
end

-- actually only used just once for playerColor
local RAID_CLASS_COLORS = RAID_CLASS_COLORS

local function GetClassColor(class)
	return format("%02X%02X%02X", RAID_CLASS_COLORS[class].r*255, RAID_CLASS_COLORS[class].g*255, RAID_CLASS_COLORS[class].b*255)
end

local function UnitLinkColor(unitflags)
	if bit_band(unitflags, COMBATLOG_OBJECT_REACTION_FRIENDLY) > 0 then
		return UnitColor[1]
	elseif bit_band(unitflags, COMBATLOG_OBJECT_REACTION_HOSTILE) > 0 then
		return UnitColor[2]
	else
		return UnitColor[3]
	end
end

-- modified from Blizzard_CombatLog
local function UnitIcon(unitFlags, reaction)
	local iconBit = bit_band(unitFlags, COMBATLOG_OBJECT_RAIDTARGET_MASK)
	if iconBit == 0 then
		return "", ""
	end
	local icon
	local iconString, braces = "", ""
	for i = 1, 8 do
		if iconBit == _G["COMBATLOG_OBJECT_RAIDTARGET"..i] then
			icon = _G["COMBATLOG_ICON_RAIDTARGET"..i]
			braces = "{"..strlower(_G["RAID_TARGET_"..i]).."}"
			break
		end
	end
	if icon then
		iconString = format(STRING_REACTION_ICON[reaction], iconBit, icon)
	end
	return iconString, braces
end

local function SpellSchool(value)
	local str = SpellSchoolString[value] or STRING_SCHOOL_UNKNOWN.." ("..value..")"
	local color = SpellSchoolColor[value] or "71D5FF"
	return "|cff"..color..str.."|r", str, color
end

local function SpellInfo(spellID, spellName, spellSchool)
	local schoolNameLocal, schoolNameChat, schoolColor = SpellSchool(spellSchool)
	local spellIcon = iconSize>1 and "|T"..select(3,GetSpellInfo(spellID))..":"..iconSize..":"..iconSize..":0:0:"..cropped.."|t" or ""
	local spellLinkLocal = format("|cff%s"..TEXT_MODE_A_STRING_SPELL.."|r", schoolColor, spellID, "", "["..spellName.."]")..spellIcon
	return schoolNameLocal, schoolNameChat, spellLinkLocal, GetSpellLink(spellID)
end

local function ResultString(schoolNameLocal, schoolNameChat, amount, overkill, critical, glancing, crushing)
	local resultStr = ""
	local resultStringLocal, resultStringChat = "", ""

	if overkill and overkillFormat then
		resultStr = resultStr.." "..format(TEXT_MODE_A_STRING_RESULT_OVERKILLING, overkill)
	end
	if critical and criticalFormat then
		resultStr = resultStr.." "..TEXT_MODE_A_STRING_RESULT_CRITICAL
	end
	if glancing and glancingFormat then
		resultStr = resultStr.." "..TEXT_MODE_A_STRING_RESULT_GLANCING 
	end
	if crushing and crushingFormat then
		resultStr = resultStr.." "..TEXT_MODE_A_STRING_RESULT_CRUSHING 
	end
	if amount then
		resultStringLocal = " "..amount.." "..schoolNameLocal..resultStr
		resultStringChat = " "..amount.." "..schoolNameChat..resultStr
	end

	return resultStringLocal, resultStringChat
end

local InterruptCheck = {}

function KCL:IneffectiveInterrupt(timestamp, subevent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellID, spellName, spellSchool)
	if subevent == "SPELL_CAST_SUCCESS" then
		InterruptCheck[sourceGUID] = true
		subevent = "CHECK_INEFFECTIVE_INTERRUPT"
		-- SPELL_INTERRUPT can have around 0.4~ sec delay/lag (in most cases)
		self:ScheduleTimer(function() self:IneffectiveInterrupt(timestamp, subevent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellID, spellName, spellSchool) end, 0.4)
	elseif subevent == "SPELL_INTERRUPT" or subevent == "SPELL_MISSED" then
		InterruptCheck[sourceGUID] = false
	elseif subevent == "CHECK_INEFFECTIVE_INTERRUPT" then
		if InterruptCheck[sourceGUID] then
			self:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, "SPELL_INEFFECTIVE_INTERRUPT", hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellID, spellName, spellSchool)
		end
		InterruptCheck[sourceGUID] = false
	end
end

local Time
local f = CreateFrame("Frame")
f:SetScript("OnUpdate", function() Time = time() end)

local lastDeath
local playerColor = GetClassColor(player.class)

function KCL:COMBAT_LOG_EVENT_UNFILTERED(event, ...)

	local timestamp, subevent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = ...

	local isDamageEvent = strfind(subevent, "DAMAGE")
	local isHealEvent = strfind(subevent, "HEAL")
	local sourceType = tonumber(strsub(sourceGUID, 5, 5))
	local destType = tonumber(strsub(destGUID, 5, 5))

	---------------
	--- Filters ---
	---------------

	if sourceName == player.name or destName == player.name then
		if not filterSelf then return end
	else
		if not filterEverythingElse then return end
	end

	-- only filter against the source unit for death events
	if (sourceType == 0 and not isDamageEvent) or (destType == 0 and isDamageEvent) then
		if not filterPlayers then return end
	else
		if not filterMonsters then return end
	end

	----------------------
	--- Unit Link/Icon ---
	----------------------
	
	local sourceNameTrim, destNameTrim = "", ""

	local sourceUnitLocal, destUnitLocal = "", ""
	local sourceUnitChat, destUnitChat = "", ""

	local sourceIconLocal, sourceIconChat = UnitIcon(sourceRaidFlags, 1)
	local destIconLocal, destIconChat = UnitIcon(destRaidFlags, 2)

	if sourceName then
		-- trim out realm name
		sourceNameTrim = strmatch(sourceName, "([^%-]+)%-?.*")
		local name, color = sourceNameTrim
		if sourceName == player.name then
			color, name = playerColor, UNIT_YOU_SOURCE
		elseif UnitInParty(sourceName) or UnitInRaid(sourceName) then
			color = ClassColor[select(2,UnitClass(sourceName))]
		elseif sourceType == 0 and (sourceName == UnitName("target") or sourceName == UnitName("focus")) then
			if sourceName == UnitName("target") then
				color = ClassColor[select(2,UnitClass("target"))]
			elseif sourceName == UnitName("focus") then
				color = ClassColor[select(2,UnitClass("focus"))]
			end
		else
			color = UnitLinkColor(sourceFlags)
		end
		sourceUnitLocal = format("|cff%s"..TEXT_MODE_A_STRING_SOURCE_UNIT.."|r", color, sourceIconLocal, sourceGUID, "["..sourceNameTrim.."]", "["..name.."]")
		sourceUnitChat = sourceIconChat.."["..sourceNameTrim.."]"
	end
	if destName then
		destNameTrim = strmatch(destName, "([^%-]+)%-?.*")
		local name, color = destNameTrim
		if destName == player.name then
			color, name = playerColor, UNIT_YOU_DEST
		elseif UnitInParty(destName) or UnitInRaid(destName) then
			color = ClassColor[select(2,UnitClass(destName))]
		elseif destType == 0 and (destName == UnitName("target") or destName == UnitName("focus")) then
			if destName == UnitName("target") then
				color = ClassColor[select(2,UnitClass("target"))]
			elseif destName == UnitName("focus") then
				color = ClassColor[select(2,UnitClass("focus"))]
			end
		else
			color = UnitLinkColor(destFlags)
		end
		destUnitLocal = format("|cff%s"..TEXT_MODE_A_STRING_DEST_UNIT.."|r", color, destIconLocal, destGUID, "["..destNameTrim.."]", "["..(destName==sourceName and QUICKBUTTON_NAME_SELF or name).."]")
		destUnitChat = destIconChat.."["..destNameTrim.."]"
	end

	local schoolNameLocal, extraschoolNameLocal = "", ""
	local schoolNameChat, extraSchoolNameChat = "", ""

	local spellIcon, extraSpellIcon
	local spellLinkLocal, extraSpellLinkLocal = "", ""
	local spellLinkChat, extraSpellLinkChat = "", ""

	----------------
	--- Suffixes ---
	----------------

	local spellID, spellName, spellSchool
	local SuffixParam1, SuffixParam2, SuffixParam3, SuffixParam4, SuffixParam5, SuffixParam6, SuffixParam7, SuffixParam8, SuffixParam9

	local prefix = strsub(subevent, 1, 5)
	if prefix == "SWING" then
		SuffixParam1, SuffixParam2, SuffixParam3, SuffixParam4, SuffixParam5, SuffixParam6, SuffixParam7, SuffixParam8, SuffixParam9 = select(12, ...)
	elseif prefix == "SPELL" or prefix == "RANGE" or prefix == "DAMAG" then
		spellID, spellName, spellSchool, SuffixParam1, SuffixParam2, SuffixParam3, SuffixParam4, SuffixParam5, SuffixParam6, SuffixParam7, SuffixParam8, SuffixParam9 = select(12, ...)

	-----------------------
	--- Spell Link/Icon ---
	-----------------------

		schoolNameLocal, schoolNameChat, spellLinkLocal, spellLinkChat = SpellInfo(spellID, spellName, spellSchool)
		if extraSpellEvent[subevent] then
			extraschoolNameLocal, extraSchoolNameChat, extraSpellLinkLocal, extraSpellLinkChat = SpellInfo(SuffixParam1, SuffixParam2, SuffixParam3)
		end
	end

	-----------------
	--- Timestamp ---
	-----------------

	local timestampLocal, timestampChat = "", ""
	if Timestamp then
		local Date = date("[%H:%M:%S]")
		timestampLocal = "|cffA9A9A9"..Date.."|r "
		timestampChat = Date.." "
	end

	---------------------
	--- Damage String ---
	---------------------

	local resultStringLocal, resultStringChat = "", ""
	if isDamageEvent or isHealEvent then
		if subevent == "SWING_DAMAGE" then
			schoolNameLocal = "|cff"..SpellSchoolColor[0x1]..ACTION_SWING.."|r"
			schoolNameChat = ACTION_SWING
		end
		resultStringLocal, resultStringChat = ResultString(schoolNameLocal, schoolNameChat, SuffixParam1, SuffixParam2, SuffixParam7, SuffixParam8, SuffixParam9)
	end

	if instanceTypeFilter then

		local isTank = (UnitGroupRolesAssigned(sourceName) == "TANK")
		local textLocal, textChat

	-----------------
	--- Subevents ---
	-----------------

		-- Deaths
		if profile.Death or profile.ChatDeath then
			-- Instagibs
			if subevent == "SPELL_INSTAKILL" then
				textLocal = profile.Death and destUnitLocal.." |cff"..EventColor["DEATH"]..ACTION_UNIT_DIED.."|r "..sourceUnitLocal..spellLinkLocal
				textChat = ChatFilter("Death") and destUnitChat.." "..ACTION_UNIT_DIED.." "..sourceUnitChat..spellLinkChat
			-- Environmental Deaths
			-- problem: the overkill parameter is always stuck on zero
			elseif subevent == "ENVIRONMENTAL_DAMAGE" then
				local environmentalType, SuffixParam1, _, SuffixParam3 = select(12, ...)
				if UnitHealth(destName) and UnitHealth(destName) == 1 then
					textLocal = profile.Death and destUnitLocal.." |cff"..EventColor["DEATH"]..ACTION_UNIT_DIED.."|r "..SuffixParam1.." ("..environmentalDamageType[environmentalType] or environmentalType..")"
					textChat = ChatFilter("Death") and destUnitChat.." "..ACTION_UNIT_DIED.." "..SuffixParam1.." ("..environmentalDamageType[environmentalType] or environmentalType..")"
				end
			-- sanity check for SuffixParam2/Overkill; couldn't find out which subevent I'm handling wrong :(
			elseif isDamageEvent and subevent ~= "ENVIRONMENTAL_DAMAGE" and subevent ~= "DAMAGE_SHIELD_MISSED" and type(SuffixParam2) == "number" and SuffixParam2 > 0 and (Time > (cd.death or 0) or destName ~= lastDeath) then
				cd.death = Time + 0.1; lastDeath = destName
				if subevent == "SWING_DAMAGE" then
					textLocal = profile.Death and destUnitLocal.." |cff"..EventColor["DEATH"]..ACTION_UNIT_DIED.."|r "..sourceUnitLocal
					textChat = ChatFilter("Death") and destUnitChat.." "..ACTION_UNIT_DIED.." "..sourceUnitChat
				else
					textLocal = profile.Death and destUnitLocal.." |cff"..EventColor["DEATH"]..ACTION_UNIT_DIED.."|r "..sourceUnitLocal..spellLinkLocal
					textChat = ChatFilter("Death") and destUnitChat.." "..ACTION_UNIT_DIED.." "..sourceUnitChat..spellLinkChat
				end
			end
		end
		-- Taunts
		if (profile.Taunt or profile.ChatTaunt) and ((profile.TankTaunt and isTank) or (not profile.TankTaunt and not isTank)) then
			if destType >= 3 and Taunt[spellID] then
				if subevent == "SPELL_AURA_APPLIED" then
					textLocal = profile.Taunt and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["TAUNT"].."taunted|r "..destUnitLocal
					textChat = ChatFilter("Taunt") and sourceUnitChat..spellLinkChat.." taunted "..destUnitChat
				end
			elseif subevent == "SPELL_CAST_SUCCESS" then
				-- AoE Taunt: Warrior 1161[Challenging Shout], Druid 5209[Challenging Roar]
				if spellID == 1161 or spellID == 5209 then
					textLocal = profile.Taunt and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["TAUNT"].."AoE "..GetSpellInfo(355).."|r"
					textChat = ChatFilter("Taunt") and sourceUnitChat..spellLinkChat.." AoE "..GetSpellInfo(355)
				-- Pet 2649[Growl], Voidwalker 3716[Torment]
				elseif profile.PetGrowl and destType >= 3 and (spellID == 2649 or spellID == 3716) then
					textLocal = profile.Taunt and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["TAUNT"].."growled|r "..destUnitLocal
					textChat = ChatFilter("Taunt") and sourceUnitChat..spellLinkChat.." growled "..destUnitChat
				end
			end
		end
		-- CC Breaker
		if (profile.CC_Break or profile.ChatCC_Break) and ((profile.TankBreak and isTank) or (not profile.TankBreak and not isTank)) then
			if subevent == "SPELL_AURA_BROKEN_SPELL" then -- broken by a spell
				textLocal = profile.CC_Break and sourceUnitLocal..extraSpellLinkLocal.." |cff"..EventColor["CC_BREAK"]..ACTION_SPELL_AURA_BROKEN.."|r "..spellLinkLocal.." on "..destUnitLocal
				textChat = ChatFilter("CC_Break") and sourceUnitChat..extraSpellLinkChat.." "..ACTION_SPELL_AURA_BROKEN.." "..spellLinkChat.." on "..destUnitChat
			elseif subevent == "SPELL_AURA_BROKEN" then -- not broken by a spell; this subevent does not seem to fire for CLEU; 4.0.1 bug?
				if sourceName then
					textLocal = profile.CC_Break and sourceUnitLocal.." |cff"..EventColor["CC_BREAK"]..ACTION_SPELL_AURA_BROKEN.."|r "..spellLinkLocal.." on "..destUnitLocal
					textChat = ChatFilter("CC_Break") and sourceUnitChat.." "..ACTION_SPELL_AURA_BROKEN.." "..spellLinkChat.." on "..destUnitChat
				else
					textLocal = profile.CC_Break and spellLinkLocal.." on "..destUnitLocal.." |cff"..EventColor["CC_BREAK"].."broken|r"
					textChat = ChatFilter("CC_Break") and spellLinkChat.." on "..destUnitChat.." broken"
				end
			end
		end
		-- Dispels / Spellsteals
		if subevent == "SPELL_DISPEL" then
			local dispelString
			if (bit_band(sourceFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) > 0 and bit_band(destFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) > 0) or
			(bit_band(sourceFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) > 0 and bit_band(destFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) > 0) then
				dispelString = profile.FriendlyDispel and ACTION_SPELL_DISPEL_DEBUFF
			else
				dispelString = profile.HostileDispel and ACTION_SPELL_DISPEL_BUFF
			end
			if dispelString then
				textLocal = profile.Dispel and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["DISPEL"]..dispelString.."|r "..destUnitLocal..extraSpellLinkLocal
				textChat = ChatFilter("Dispel") and sourceUnitChat..spellLinkChat.." "..dispelString.." "..destUnitChat..extraSpellLinkChat
			end
		elseif subevent == "SPELL_STOLEN" then
			textLocal = profile.Dispel and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["DISPEL"]..ACTION_SPELL_STOLEN.."|r "..destUnitLocal..extraSpellLinkLocal
			textChat = ChatFilter("Dispel") and sourceUnitChat..spellLinkChat.." "..ACTION_SPELL_STOLEN.." "..destUnitChat..extraSpellLinkChat
		-- Reflects / Misses
		elseif subevent == "SPELL_MISSED" then
			if SuffixParam1 == "REFLECT" then
				textLocal = profile.Reflect and destUnitLocal.." |cff"..EventColor["REFLECT"]..ACTION_SPELL_MISSED_REFLECT.."|r "..sourceUnitLocal..spellLinkLocal
				textChat = ChatFilter("Reflect") and destUnitChat.." "..ACTION_SPELL_MISSED_REFLECT.." "..sourceUnitChat..spellLinkChat
			-- bit confused how miss stuff should be handled,
			-- since there aren't local/chat option filters for it ..
			-- using the normal options instead as a local/chat filter
			elseif profile.MissAll or (Taunt[spellID] and profile.MissTaunt) or (Interrupt[spellID] and profile.MissInterrupt) or (CrowdControl[spellID] and profile.MissCC) then
				local taunt, interrupt, cc = Taunt[spellID], Interrupt[spellID], CrowdControl[spellID]
				if (taunt and profile.Taunt) or (interrupt and profile.Interrupt) or (cc and profile.CrowdControl) then
					textLocal = sourceUnitLocal..spellLinkLocal.." on "..destUnitLocal.." |cffFF7800"..ACTION_SPELL_CAST_FAILED.."|r ("..missType[SuffixParam1]..")"
				end
				if (taunt and profile.ChatTaunt) or (interrupt and profile.ChatInterrupt) or (cc and profile.ChatCrowdControl) then
					textChat = sourceUnitChat..spellLinkChat.." on "..destUnitChat.." "..ACTION_SPELL_CAST_FAILED.." ("..missType[SuffixParam1]..")"
				end
				-- check if the interrupt didn't miss, instead of being wasted
				if profile.WastedInterrupt and Interrupt[spellID] then
					self:IneffectiveInterrupt(...)
				end
			end
		-- Death Prevents; -- Priest 48153[Guardian Spirit], Paladin 66235[Ardent Defender]
		elseif subevent == "SPELL_HEAL" and (spellID == 48153 or spellID == 66235) then
			textLocal = profile.DeathPrevent and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["RESURRECTION"].."healed "..destUnitLocal
			textChat = ChatFilter("DeathPrevent") and sourceUnitChat..spellLinkChat.." healed "..destUnitChat..spellLinkChat
		-- Resurrections
		elseif subevent == "SPELL_RESURRECT" then
			textLocal = profile.Resurrection and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["RESURRECTION"]..ACTION_SPELL_RESURRECT.."|r "..destUnitLocal
			textChat = ChatFilter("Resurrection") and sourceUnitChat..spellLinkChat.." "..ACTION_SPELL_RESURRECT.." "..destUnitChat
		-- Crowd Control
		elseif subevent == "SPELL_AURA_APPLIED" and CrowdControl[spellID] then
			textLocal = profile.CrowdControl and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["CROWDCONTROL"].."CC'ed|r "..destUnitLocal
			textChat = ChatFilter("CrowdControl") and sourceUnitChat..spellLinkChat.." CC'ed "..destUnitChat
		-- "pre-Interrupts"
		elseif subevent == "SPELL_CAST_SUCCESS" and Interrupt[spellID] then
			self:IneffectiveInterrupt(...)
		-- Interrupts
		elseif subevent == "SPELL_INTERRUPT" then
			textLocal = profile.Interrupt and sourceUnitLocal..spellLinkLocal.." |cff"..EventColor["INTERRUPT"]..ACTION_SPELL_INTERRUPT.."|r "..destUnitLocal..extraSpellLinkLocal
			textChat = ChatFilter("Interrupt") and sourceUnitChat..spellLinkChat.." "..ACTION_SPELL_INTERRUPT.." "..destUnitChat..extraSpellLinkChat
			-- confirm interrupt
			self:IneffectiveInterrupt(...)
		-- Wasted Interrupt
		elseif subevent == "SPELL_INEFFECTIVE_INTERRUPT" then
			textLocal = profile.WastedInterrupt and sourceUnitLocal.." |cffFF7800wasted|r "..spellLinkLocal.."  on "..destUnitLocal
			textChat = ChatFilter("WastedInterrupt") and sourceUnitChat.." wasted "..spellLinkChat.." on "..destUnitChat
		end

--[[
		-- Feast, Repair Bot, Fun
		if profile.Feast then
			if subevent == "SPELL_CREATE" and
			(spellID == 87644 or -- [Seafood Magnifique Feast]
			spellID == 87643 or -- [Broiled Dragon Feast]
			spellID == 87915 or -- [Goblin Barbecue Feast]
			spellID == 57426 or -- [Fish Feast]
			spellID == 57301 or -- [Great Feast]
			spellID == 56245 or -- [Chocolate Celebration Cake]
			spellID == 57301) then -- [Lovely Cake]
				local timeBG = GetBattlefieldInstanceRunTime()
				if instanceType == "pvp" and timeBG ~= 0 and timeBG < 130000 then text = "["..sourceNameTrim.."] prepared a "..spellLinkChat
				else
					textLocal = sourceUnitLocal.." prepared a "..spellLinkLocal
					textChat = timestampChat..sourceIconChat.."["..sourceNameTrim.."] prepared a "..spellLinkChat
				end
			end
		end
		if profile.RepairBot then
			if (subevent == "SPELL_CREATE" and spellID == 51710) or
			(subevent == "SPELL_SUMMON" and spellID == 67826 or spellID == 54711) then
				textLocal = sourceUnitLocal.." used "..spellLinkLocal
				textChat = timestampChat..sourceIconChat.."["..sourceNameTrim.."] used "..spellLinkChat
			end
		end
		if profile.Fun and player.name ~= sourceName then -- dont show yourself xD
			if (subevent == "SPELL_CAST_START" and spellID == 58493) or -- S 58493[Mohawked!]
			(subevent == "SPELL_AURA_APPLIED" and
			(spellID == 61834 or -- [Manabonked!] (NPC)
			spellID == 61819 or -- [Manabonked!] (Item)
			spellID == 60122 or -- [Baby Spice]
			((spellID == 55346 or -- [Dark Jade Beam]
			spellID == 56190 or -- [Shadow Crystal Focusing Lens]
			spellID == 56191) and destName ~= sourceName) or -- [Shadow Jade Focusing Lens]
			strfind(spellName, "Costume") or
			spellID == 27571 or -- [Cascade of Roses]  Love is in the Air
			spellID == 61815 or -- [Sprung!]  Noblegarden
			spellID == 44212 or -- [Jack-o'-Lanterned!]  Hallow's End
			spellID == 61781 or -- [Turkey Feathers]  Pilgrim's Bounty
			spellID == 44755)) or -- [Snowflakes]  Feast of Winter Veil
			(subevent == "SPELL_CREATE" and
			(spellID == 61031 or -- [Toy Train Set]
			spellID == 49844 or -- [Using Direbrew's Remote]
			spellID == 43808 or -- [Brewfest Pony Keg]
			spellID == 62857)) then -- [Sandbox Tiger]
				if spellID == 61031 or spellID == 49844 then text = "["..sourceNameTrim.."] used "..spellLinkChat -- Toy Train, Direbrew's Remote
				else
					if destName then a = " on "..destUnitLocal else a = "" end text = sourceUnitLocal.." used "..spellLinkLocal..a
				end
			end
		end
		-- Class Ability
		if subevent == "SPELL_AURA_APPLIED" then
			if ((profile.abilityExcludeSelf and sourceName ~= player.name) or not profile.abilityExcludeSelf) and
			((spellID == 20707 and profile.abilitySoulstone) or
			(spellID == 54646 and profile.abilityFocusMagic) or
			(spellID == 50720 and profile.abilityVigilance) or
			(spellID == 29166 and profile.abilityInnervate) or
			(spellID == 6346 and profile.abilityFearWard) or
			(spellID == 2096 and profile.abilityMindVision and destName ~= sourceName)) then
				textLocal = sourceUnitLocal.." cast "..spellLinkLocal.." on "..destUnitLocal
				if spellID ~= 2096 then text = timestampChat..sourceIconChat.."["..sourceNameTrim.."] cast "..spellLinkChat.." on "..destUnitChat
			end
		-- quick and dirty [Misdirection] and [Tricks of the Trade] fix
		elseif subevent == "SPELL_CAST_SUCCESS" then
			if ((profile.abilityExcludeSelf and sourceName ~= player.name) or not profile.abilityExcludeSelf) and
				((spellID == 34477 and profile.abilityMisdirection and destName ~= sourceName) or
				(spellID == 57934 and profile.abilityTricksTrade and destName ~= sourceName)) then
				textLocal = sourceUnitLocal.." cast "..spellLinkLocal.." on "..destUnitLocal
				if spellID ~= 2096 then text = timestampChat..sourceIconChat.."["..sourceNameTrim.."] cast "..spellLinkChat.." on "..destUnitChat
			end
		end
		if subevent == "SPELL_CAST_SUCCESS" then
			if profile.abilityBloodlust and ((spellID == 32182 or spellID == 2825 or spellID == 80353) and Time > (cd.Bloodlust or 0)) or
			(spellID == 43987 and profile.abilityRitualRefreshment and Time > (cd.RitualRefreshment or 0)) then
				if spellID == 32182 or spellID == 2825 or spellID == 80353 then cd.Bloodlust = Time + 130 -- set cooldowns
				elseif spellID == 43987 then cd.RitualRefreshment = Time + 130 end
				local timeBG = GetBattlefieldInstanceRunTime()
				if instanceType == "pvp" and timeBG ~= 0 and timeBG < 130000 then text = "["..sourceNameTrim.."] cast "..spellLinkChat -- Heroism/Bloodlust/Timewarp "wasted" before bg starts; Ritual of Refreshment
					chatType = "BATTLEGROUND"
				else
					textLocal = sourceUnitLocal.." cast "..spellLinkLocal
					textChat = timestampChat..sourceIconChat.."["..sourceNameTrim.."] cast "..spellLinkChat
				end
			end
		end
]]

	--------------
	--- Output ---
	--------------
	
	--	if textChat then print(textChat) end

		if textLocal or textChat then
			if textLocal and chatWindow > 1 then
				textLocal = timestampLocal..textLocal..resultStringLocal
				ChatFrame:AddMessage(textLocal)
			end
			if textChat and chatChannel > 1 then
				textChat = timestampChat..textChat..resultStringChat
				SendChatMessage(textChat, chatType, nil, channel)
			end
			-- LibSink can't handle (item/spell) hyperlinks
			-- dirty hack + ugly pattern
			if textChat and profile.sink20OutputSink == "Channel" then
				self:Pour(gsub(textChat, "|c.-|r", strmatch(textChat, "|h(%[.+%])|h") or ""))
			elseif textLocal and profile.sink20OutputSink ~= "Channel" then
				self:Pour(textLocal)
			end
		end
	end
end